#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import json
import os
import numpy as np
import pickle
from scipy.interpolate import CubicSpline
from copy import deepcopy

# Own modules
from PE_CBL_models import CBL
from observations import *

# TODO: comprehensive documentation is missing.

# TODO: code should already permit to perturb and/or estimate multiple parameters
#  But it needs testing.

# TODO: code should already permit assimilating u and v observations.
#  But it needs testing.

# TODO: add other options for variance inflation (multiplicative, additive).

# TODO: add other options for parameter transformations.

verbose = True

def observation_operator(fp,xp,x):
    f = np.interp(x, xp, fp)
    return f

def gaspari_cohn(r,cutoff):
    """Gaspari-Cohn function.
    https://github.com/ecosme38/Data-Assimilation-Notebooks/blob/master/Illustrate-Gaspari-Cohn-function-and-covariance-localization.ipynb
    """
    r = r.astype('float')
    ra = np.abs(r/cutoff)
    gp = np.zeros_like(r)
    
    i = (ra<=1.)
    gp[i] = -0.25*ra[i]**5+0.5*ra[i]**4+0.625*ra[i]**3-5./3.*ra[i]**2+1.
    
    i = (ra>1)&(ra<=2.)
    gp[i] = 1./12.*ra[i]**5-0.5*ra[i]**4+0.625*ra[i]**3+5./3.*ra[i]**2-5.*ra[i]+4.-2./3./ra[i]
    
    return gp

def compute_distances(xcoords,ocoords):
    # Number of state vector elements (coordinates are given)
    nstate = xcoords.size

    # Number of observations (coordinates are given)
    nobs = ocoords.size

    # Preallocate matrix to store distances between
    # pairs of state vector elements and observations
    dist = np.zeros((nstate,nobs)) + np.nan
    
    # i loops over state vector elements
    # j loops over observations
    for i in range(nstate):
        for j in range(nobs):
            dist[i,j] = np.abs(xcoords[i]-ocoords[j])
    
    # Return distances
    return dist

def eakf(xb,o,var_o,xcoord,ocoord,dist,\
                    localization_cutoff=np.infty,return_covariances_increments_and_innovations=False):

    def increments_in_obs_space(y_p,o,var_o):
    
        # Compute ensemble mean, ensemble variance and weight of prior
        mean_p = np.mean(y_p)
        var_p = np.var(y_p)
        weight_p = 1./var_p

        # Observation weight
        weight_o = 1./var_o
    
        # Compute weight of posterior, then ensemble variance of posterior
        weight_u =  weight_o + weight_p
        var_u = 1./weight_u

        # Shift the mean from mean_p to mean_u
        mean_u = (mean_p*weight_p + o*weight_o)*var_u 
        
        # Compute ensemble of analysis increments in observation space.
        # There are three terms in delta_y; the first two are the posterior
        # ensemble state (shifted mean + adjusted perturbations; e.g. Eq 19 in Anderson 2003);
        # the last one (y_p) is the prior state.
        # Their difference is the increment in observation space.
        alpha = (var_u*weight_p)**0.5
        #delta_y = mean_u*np.ones(nens) + (alpha-1)*y_p - alpha*mean_p*np.ones(nens) 
        delta_y = mean_u*np.ones(nens) + alpha*(y_p - mean_p*np.ones(nens)) - y_p

        return delta_y,weight_p
    
    # Read out relevant dimensions
    nvar = xb.shape[0]
    nens = xb.shape[1]
    nobs = o.shape[0]

    # Make a copy of the input array. Necessary in order to store correctly
    # the last forecast and the analysis.
    xu = np.copy(xb)

    # Compute unique values of distances and localization function.
    # This spares some computations because distances are discrete,
    # so a given value occurs more than once.
    ds = np.sort(np.unique(dist))
    loc = gaspari_cohn(ds,cutoff=localization_cutoff)

    # Preallocate array for model equivalents
    yp = np.zeros((nobs,nens))+np.nan
    increments = np.zeros((nvar,nobs))+np.nan

    # Preallocate array to save variances and covariances
    # (they are adjusted after ingesting every single observation).
    # This is not needed by the algorithm, but may be useful
    # for analysis purposes.
    if return_covariances_increments_and_innovations:
        cov_xx = np.zeros((nobs,nvar))+np.nan
        cov_xy = np.zeros((nobs,nvar))+np.nan
        cov_yy = np.zeros((nobs))+np.nan
        innovations_mean = np.zeros((nobs))+np.nan
        increments_mean = np.zeros((nobs,nvar))+np.nan
    
    # Process observations sequentially
    for i in range(nobs):

        # Apply observation operator to prior ensemble, i.e.
        # compute model equivalents of observations
        for j in range(nens):
            yp[i,j] = observation_operator(xu[:,j],xcoord,ocoord[i])

        # Store innovation wrt to ensemble mean
        if return_covariances_increments_and_innovations:
            innovations_mean[i] = o[i]-yp[i,:].mean()

        # Determine analysis increments in observation space
        # weight_p is a scalar
        # delta_y has dimension nens
        delta_y,weight_p = increments_in_obs_space(yp[i,:],o[i],var_o[i])

        # Update state variables in each ensemble member using ensemble covariances
        # The default in np.cov is N-1 degrees of freedom.
        for k in range(nvar):
            cov = np.cov(xu[k,:],yp[i,:]) # scalar
            # Compute analysis increments (=adjustments of ensemble perturbations)
            increments = cov[0,1]*weight_p*delta_y*\
                np.interp(dist[k,i],ds,loc) # nens values; second row is covariance localization
            xu[k,:] = xu[k,:] + increments
            if return_covariances_increments_and_innovations:
                cov_xx[i,k] = cov[0,0]
                cov_xy[i,k] = cov[0,1]
                cov_yy[i] = cov[1,1]
                increments_mean[i,k] = increments.mean()


    # Return
    if return_covariances_increments_and_innovations:
        return xu,yp,cov_xx,cov_xy,cov_yy,innovations_mean,increments_mean
    else:
        return xu,yp

def letkf(xb,o_global,var_o,xcoord,ocoord,dist,\
                     localization_cutoff=np.infty,inflation_rho=1.00):
    # Default size of local domain is set to 6 (=13 points), following Baek et al.    
    # Following strictly http://dx.doi.org/10.1016/j.physd.2006.11.008
    assert inflation_rho >= 1, \
        f"LETKF: Inflation factor must be > 1, got: {inflation_rho}"

    # Read out relevant dimensions
    nvar = xb.shape[0]
    nens = xb.shape[1]
    nobs = o_global.shape[0]

    # Create an empty array for the analysis.
    # Necessary in order to store correctly
    # the last forecast and the analysis.
    xa = np.zeros(xb.shape)+np.nan

    # Apply observation operator to prior ensemble, i.e.
    # compute model equivalents of observations
    yb_global = np.zeros((nobs,nens))+np.nan
    for i in range(nobs):
        for j in range(nens):
            yb_global[i,j] = observation_operator(xb[:,j],xcoord,ocoord[i])

    # Compute ensemble mean in observation space
    ybm_global = np.mean(yb_global,axis=1)

    # Compute matrix of ensemble perturbations in observation space
    Yb_global = yb_global-np.tile(ybm_global,(nens,1)).T

    # Compute ensemble mean in state space
    xbm = np.mean(xb,axis=1)

    # Compute matrix of ensemble perturbations in state space
    Xb = xb-np.tile(xbm,(nens,1)).T
    
    # Loop over grid points
    for m in range(nvar):

        # Get local model equivalents
        ybm = ybm_global[dist[m,:]<=localization_cutoff]
        Yb = Yb_global[dist[m,:]<=localization_cutoff]

        # Get local observations and their error variances
        yo = o_global[dist[m,:]<=localization_cutoff]
        vo = var_o[dist[m,:]<=localization_cutoff]

        # Form R matrix
        R = np.diag(vo)

        # Compute C matrix
        # The following statements are equivalent, the first is slightly faster
        # C = Yb.T@np.linalg.inv(R)
        # C = np.linalg.solve(R,Yb).T
        C = Yb.T@np.linalg.inv(R)

        # Compute analysis error variance matrix Pa
        Pa = np.linalg.inv( (nens-1)*np.eye(nens)/inflation_rho + C@Yb )
        
        # Compute weight matrix (symmetric square root of Pa)
        # Can use eigh and evectors.T because Pa is real symmetric
        evalues, evectors = np.linalg.eigh((nens-1)*Pa)
        Wa = evectors * np.sqrt(evalues) @ evectors.T
        
        # Compute ensemble mean weights
        wam = Pa @ C @ (yo-ybm)

        # Compute weights of ensemble perturbations
        Wai = Wa + np.tile(wam,(nens,1)).T
        
        # Compute local analysis ensemble members and store them
        # The ensemble mean would be
        # xam[m,:] = xbm[m] + Xb[m,:]@wam
        # but it's not necessary
        xa[m,:] = np.tile(xbm[m],(nens,1)).T + Xb[m,:] @ Wai
        
    return xa,yb_global

class cycle:
    def __init__(self,\
                 cbl_settings,\
                 nr,\
                 nens,\
                 FILTER,\
                 trun,\
                 tspinup_assim,\
                 assimilation_interval,\
                 ocoord,\
                 observations,\
                 obs_error_sdev_assimilate,\
                 localization_cutoff,\
                 inflation_rtps_alpha,\
                 simulate_underdispersiveness,\
                 rnseed):

        # Sanity check
        assert trun>tspinup_assim, \
            f"Pre-assimilation spinup time must be shorter than run length, got: {trun, tspinup_assim}"

        # How many cycles do you need to run?
        ncycles = int(trun/assimilation_interval)
        time = (np.arange(ncycles)+1)*assimilation_interval

        # Do you want to have a spinup time for the assimilation?
        # I.e., an interval between ensemble initialization
        # and the first assimilation
        nspinupcycles = int(tspinup_assim/assimilation_interval)
        analysis_times = np.arange(nspinupcycles,ncycles)

        # Define state vector length
        state_vector_length = nr.x0.size
        initial_state_vector = nr.x0
        xcoord = nr.state_coordinates
        nobs = observations.shape[1]
        npar = cbl_settings['parameter_number']

        # Special case: if you re-use a nature run in a experiment without
        # parameter estimation, the extra elements in the state vector must
        # be stripped, otherwise the ensemble members won't be initialized
        # properly
        if hasattr(nr,"has_parameters_in_state") and cbl_settings['do_parameter_estimation'] == False:
            print("Warning: you are re-using a nature run with augmented state in a run without parameter estimation")
            state_vector_length -= nr.parameter_number
            initial_state_vector = np.delete(nr.x0,-nr.parameter_number,0)
            xcoord = np.delete(nr.state_coordinates,-nr.parameter_number)

        # Preallocate arrays
        initial_state     = np.zeros((1,state_vector_length,nens)) + np.nan
        backgrounds       = np.zeros((ncycles,state_vector_length,nens)) + np.nan
        analyses          = np.zeros((ncycles,state_vector_length,nens)) + np.nan
        model_equivalents = np.zeros((ncycles,nobs,nens)) + np.nan
        if cbl_settings['return_covariances_increments_and_innovations']:
            cov_py = np.zeros((ncycles,nobs,npar)) + np.nan
            cov_pp = np.zeros((ncycles,nobs,npar)) + np.nan
            cov_yy = np.zeros((ncycles,nobs)) + np.nan
            innov = np.zeros((ncycles,nobs)) + np.nan
            increments = np.zeros((ncycles,nobs,npar)) + np.nan

        # Turn inflation coefficients into an array
        inflation_coefficients_rtps = np.ones(state_vector_length)*inflation_rtps_alpha

        # Initialize ensemble
        # Get the right model class from the nature run.
        # The class must have "initialize", "run" and "update" methods.
        if isinstance(nr,CBL):
            # Initialize assimilation ensemble by adding
            # random perturbations to initial state of nature run.
            model = CBL(cbl_settings)
            x0 = np.tile(initial_state_vector,(nens,1)).T
            model.initialize(x0,coordinates=xcoord)
            # In case of parameter estimation, update variance inflation settings
            if model.do_parameter_estimation:
                inflation_coefficients_rtps[-model.parameter_number:] = model.parameter_inflation_rtps_alpha

        # Store initial ensemble state
        initial_state[0,:,:] = model.x0[:,:]

        # If array of inflation coefficients is filled with zeros,
        # turn off posterior variance inflation
        if np.sum(inflation_coefficients_rtps) == 0.0:
            inflate_rtps = False
        else:
            inflate_rtps = True

        history = []
        # Start cycling
        for k in range(ncycles):
            if verbose: print('Cycle %04u    : '%k,end='')

            # Integrate up to analysis time
            model.maxtime = assimilation_interval
            output_full_history = False
            if output_full_history:
                model.run(output_full_history=True)
                history.append(model.history)
            else:
                model.run()

            # If you want to simulate ensemble underdispersiveness
            # (=spread smaller than actual error of the ensemble mean),
            # perturb the ensemble mean before the analysis.
            # Only for the state, not for the parameter!
            if simulate_underdispersiveness:
                # Ensemble mean repeated nens times
                ensmean = model.x[:-model.parameter_number,:].mean(axis=1)
                xbm = np.tile(ensmean,(nens,1))
                # Ensemble perturbations
                xbp = model.x[:-model.parameter_number,:].T-xbm
                # Define positions of random perturbations
                # enforce two at boundaries to avoid extrapolation
                randomsize = model.perturbations_smooth_number
                ipert = np.arange(0,randomsize-1)*(model.nz//randomsize)
                ipert = np.append(ipert,model.nz-1)
                # Generate random perturbations
                # Change the seed every time you go through this,
                # otherwise the perturbations are always the same;
                # and you have bias instead of randomness
                RNG = np.random.default_rng(seed=rnseed+k*10000)
                pert_random = RNG.standard_normal(randomsize)
                f = CubicSpline(ipert,pert_random)
                pert_interpolated = f(np.arange(model.nz))*model.error_growth_perturbations_amplitude
                # Perturbed ensemble mean
                xbm_new = np.tile(ensmean+pert_interpolated,(nens,1))
                # Store the perturbed ensemble mean plus original ensemble perturbations
                bkgd = np.zeros(model.x.shape)
                bkgd[:-model.parameter_number,:] = (xbm_new+xbp).T
                bkgd[-model.parameter_number:,:] = model.x[-model.parameter_number:,:]
            else:
                bkgd = model.x[:,:]

            # Store prior
            backgrounds[k,:,:] = bkgd

            if k not in analysis_times: # Do not assimilate anything at this time; simply let the ensemble run
            
                if verbose: print('~analysis skipped~')

                # Propagate the prior state as is to the next iteration
                #x0 = model.x[:,:]
                x0 = backgrounds[k,:,:]

                # Compute the corresponsing model equivalents
                for i in range(nobs):
                    for j in range(nens):
                        model_equivalents[k,i,j] = observation_operator(x0[:,j],xcoord,ocoord[k,i])

            else: # Assimilate

                # Compute prior spread (necessary for variance inflation)
                if inflate_rtps:
                    sigp = np.std(backgrounds[k,:,:],axis=1)

                # Compute distances between current observations and state vector variables
                dist = compute_distances(xcoord,ocoord[k,:])

                # Replace any NaN distances with 0
                # NaN distances only occur when the state is augmented with global
                # parameters, whose location is undefined. Setting the distance to
                # global parameters equal to zero ensures that the covariance
                # matrix for global parameters is identity, as required by
                # global updating.
                np.nan_to_num(dist,copy=False)

                # Assimilate
                if FILTER == 'EAKF':
                    if cbl_settings['do_parameter_estimation'] and cbl_settings['return_covariances_increments_and_innovations']:
                        analysis,model_equivalents[k,:,:],cov_xx,cov_xy,cov_yy_dum,innov_dum,increm_dum = eakf(backgrounds[k,:,:],observations[k,:],
                                    obs_error_sdev_assimilate,
                                    xcoord,
                                    ocoord[k,:],
                                    dist,
                                    localization_cutoff=localization_cutoff,return_covariances_increments_and_innovations=True)
                        increments[k,:,-npar] = increm_dum[:,-npar]
                        cov_pp[k,:,-npar] = cov_xx[:,-npar]
                        cov_py[k,:,-npar] = cov_xy[:,-npar]
                        cov_yy[k,:] = cov_yy_dum
                        innov[k,:] = innov_dum
                    else:
                        analysis,model_equivalents[k,:,:] = eakf(backgrounds[k,:,:],observations[k,:],
                                    obs_error_sdev_assimilate,
                                    xcoord,
                                    ocoord[k,:],
                                    dist,
                                    localization_cutoff=localization_cutoff)
                elif FILTER == 'LETKF':
                    analysis,model_equivalents[k,:,:] = letkf(backgrounds[k,:,:],observations[k,:],
                                obs_error_sdev_assimilate,
                                xcoord,
                                ocoord[k,:],
                                dist,
                                localization_cutoff=localization_cutoff,
                                inflation_rho=1.0)

                x0 = analysis

                # Posterior variance inflation (relaxation to prior spread)
                # http://dx.doi.org/10.1175/MWR-D-11-00276.1
                # Variance inflation is optional for regular state vector elements
                if inflate_rtps:
                    # Add 1e-10 to posterior variance (denominator) to avoid NaN
                    sigu = np.std(x0,axis=1)
                    inflation_factor = np.tile(inflation_coefficients_rtps*(sigp-sigu)/(sigu+1e-10),(nens,1))
                    xbm = np.tile(x0.mean(axis=1),(nens,1))
                    xbp = x0.T-xbm
                    x0 = (xbm+xbp*(1.+inflation_factor)).T

            # Store posterior
            analyses[k,:,:] = x0

            # Save initial conditions for next cycle
            model.update(x0)

        # Store the output
        self.ncycles = ncycles
        self.time = time
        self.analysis_times = analysis_times
        self.assimilation_interval = assimilation_interval
        self.initial_state = initial_state
        self.backgrounds = backgrounds
        self.analyses = analyses
        self.obs_coordinates = ocoord
        self.observations = observations
        self.model_equivalents = model_equivalents
        self.state_coordinates = xcoord
        self.obs_error_sdev_assimilate = obs_error_sdev_assimilate
        self.history = history
        self.nz = nr.nz
        self.zt = nr.zt
        self.nens = nens
        self.initial_perturbed_parameters = model.initial_perturbed_parameters
        if cbl_settings['return_covariances_increments_and_innovations']:
            self.cov_py = cov_py
            self.cov_pp = cov_pp
            self.cov_yy = cov_yy
            self.innov = innov
            self.increments = increments

class experiment:
    def __init__(self,settings):
        # Read attributes from dictionary of settings
        for k, v in settings.items():
            setattr(self, k, v)

        """
        # Create additional attributes
        self.obs_error_sdev_generate = np.ones(self.nobs)*self.obs_error_sdev_generate
        self.obs_error_sdev_assimilate = np.ones(self.nobs)*self.obs_error_sdev_assimilate
        if self.obs_kinds == 'theta':
            self.obs_kinds = np.array(['theta']*self.nobs)

        self.obs_coordinates = np.linspace(self.obs_coordinates_min,self.obs_coordinates_max,self.nobs)
        """

        # If necessary, get nature run and related information from an external file
        if self.nature_run_from_file:
            assert os.path.isfile(self.nature_run_from_file),\
                f'Nature run file {self.nature_run_from_file} does not exist'
            exp = pickle.load(open(self.nature_run_from_file, "rb"))
            self.nr = deepcopy(exp.nr)
            self.truths = deepcopy(exp.truths)
            self.observations = deepcopy(exp.observations)
            self.obs_coordinates = deepcopy(exp.obs_coordinates)
            self.obs_error_sdev_generate = np.ones(self.nobs)*self.obs_error_sdev_generate
            self.obs_error_sdev_assimilate = np.ones(self.nobs)*self.obs_error_sdev_assimilate
            if not self.do_parameter_estimation:
                self.nr.do_parameter_estimation = False

        # Get CBL settings from file
        assert os.path.isfile(self.path+self.cbl_settings_file),\
            f'CBL settings file {self.path+self.cbl_settings_file} does not exist'
        with open(self.path+self.cbl_settings_file, 'r') as fp:
            self.cbl_settings = dict(json.load(fp))

        # Propagate parameter estimation settings to CBL model
        if self.do_parameter_estimation:
            self.cbl_settings['do_parameter_estimation'] = True
        else:
            self.cbl_settings['do_parameter_estimation'] = False

        ################################################################################################
        # CASE 1: synthetic observations (OSSE with nature run)
        if not hasattr(self,'nr') and self.type == 'OSSE':

            # Create additional attributes
            self.obs_error_sdev_generate = np.ones(self.nobs)*self.obs_error_sdev_generate
            self.obs_error_sdev_assimilate = np.ones(self.nobs)*self.obs_error_sdev_assimilate
            if self.obs_kinds == 'theta':
                self.obs_kinds = np.array(['theta']*self.nobs)

            self.obs_coordinates = np.linspace(self.obs_coordinates_min,self.obs_coordinates_max,self.nobs)

            # Consistency check on observation info
            nobs = self.nobs
            if self.obs_coordinates.ndim == 1:
                assert (self.obs_coordinates.size==self.obs_kinds.size==self.obs_error_sdev_assimilate.size==self.obs_error_sdev_generate.size),\
                    f'obs_coordinates, obs_kinds and obs_error_sdev must have the same size,\
                    got {self.obs_coordinates.size,self.obs_kinds.size,self.obs_error_sdev_assimilate.size,self.obs_error_sdev_assimilate.size}'

            # Consistency check on time coordinates
            nassim = self.trun/self.assimilation_interval
            assert (nassim.is_integer()),\
                f'trun/assimilation_interval must be integer,\
                got {self.trun,self.assimilation_interval,self.trun/self.assimilation_interval}'

            # If you don't have a pre-existing nature run, you need to run it

            # Define initial conditions of nature run
            if self.tspinup > 0:
                # Create initial state with random perturbations, then integrate
                # one deterministic run for the spinup period
                # Initial condition for the nature run is the last state of the spinup run
                if verbose: print('Spinup        : ',end='')
                sp = CBL(self.cbl_settings)
                sp.maxtime = self.tspinup
                sp.initialize(1) 
                sp.run()
                initial_conditions = sp.x
                coords = sp.state_coordinates
            else:
                # No spinup, nature run gets unperturbed initial condition
                sp = CBL(self.cbl_settings)
                sp.initialize(1)
                initial_conditions = sp.x0
                coords = sp.state_coordinates

            # Nature run: initialize, then integrate
            if verbose: print('Nature run    : ',end='')
            nr = CBL(self.cbl_settings)
            nr.maxtime = self.trun
            nr.initialize(initial_conditions,coordinates=coords)
            nr.run(output_full_history=True)
            if self.cbl_settings["do_parameter_estimation"]:
                nr.has_parameters_in_state = True

            # Draw observations from nature run
            # Need nested loops because observation times (k) are independent
            # and observation error variance depends on location (kk)
            state_coordinates = nr.zt
            truths = np.zeros((int(nassim),nobs))+np.nan
            observations = np.zeros((int(nassim),nobs))+np.nan
            for i in range(int(nassim)):
                valid_time = (i+1)*self.assimilation_interval
                assert(np.mod(self.assimilation_interval,nr.dt)==0), f'Assimilation interval must be a multiple of dt, got {self.assimilation_interval,nr.dt}'
                time_index = np.argwhere(nr.history['time'] == valid_time)[0][0]
                for j in range(nobs):
                    obs_coordinate = self.obs_coordinates[j]
                    if self.obs_kinds[j]=='theta':
                        variable = nr.history['theta'][:,time_index]
                    elif self.obs_kinds[j]=='u':
                        variable = nr.history['u'][:,time_index]
                    elif self.obs_kinds[j]=='v':
                        variable = nr.history['v'][:,time_index]
                    truths[i,j] = observation_operator(variable,state_coordinates,obs_coordinate)
                    # Change the seed every time you go through this,
                    # otherwise the perturbations are always the same;
                    # and you have bias instead of randomness
                    seed = self.rnseed+j*1000+i*100000
                    RNG = np.random.default_rng(seed=seed)
                    observations[i,j] = truths[i,j]+\
                        RNG.normal(0,self.obs_error_sdev_generate[j])

            # In this case the observations coordinates are the same at all analysis times,
            # so a 1-d vector. Expand it to 2d, with the same dimensions
            # as the observations vector
            if self.obs_coordinates.ndim == 1:
                ocoords = self.obs_coordinates[None,:]+np.zeros(observations.shape)

            # Randomize order of observations
            if self.randomize_obs:
                observations,ocoords,truths = random_sort_observations(observations,ocoords,truths)

            # Store nature run, truths and observations
            self.nr = nr
            self.truths = truths
            self.observations = observations
            self.obs_coordinates = ocoords

        ################################################################################################
        # CASE 2: real (or pseudo-) observations
        if not hasattr(self,'nr') and self.type == 'real':

            # Read observations and their coordinates
            observations, ocoords = decode_observations(self.observation_files)
            assert (observations.shape == ocoords.shape),\
                f'observation and coordinate arrays must have the same size'

            # Consistency check on time coordinates
            nassim = self.trun/self.assimilation_interval
            assert (nassim.is_integer()),\
                f'trun/assimilation_interval must be integer,\
                got {self.trun,self.assimilation_interval,self.trun/self.assimilation_interval}'

            # Randomize order of observations
            if self.randomize_obs:
                observations,ocoords = random_sort_observations(observations,ocoords)

            # Throw away any observations in the spinup time before assimilations begin
            # Exclude the initial time too (so the +1 in the line below)
            nspinup = self.tspinup/self.assimilation_interval+1
            observations = observations[int(nspinup):,:]
            ocoords = ocoords[int(nspinup):,:]

            # Throw away any observations beyond tspinup+trun
            observations = observations[:int(nassim),:]
            ocoords = ocoords[:int(nassim),:]

            # There is no nature run and no truth in this case.
            # The only thing you can store is the observations.
            self.nobs = observations.shape[1]
            self.observations = observations
            self.obs_coordinates = ocoords
            self.obs_error_sdev_assimilate = np.ones(self.nobs)*self.obs_error_sdev_assimilate
            if self.obs_kinds == 'theta':
                self.obs_kinds = np.array(['theta']*self.nobs)

            # The da cycle depends on a "nr" structure to get information
            # about the model properties.
            # Depending on the configuration, the da cycle can 
            # create the initial ensemble perturbations around
            # a spun-up model state, which is stored in nr.x0.
            # So, run the forecast model for the spinup time if necessary.
            # Define initial conditions of nature run
            if self.tspinup > 0:
                # Create initial state with random perturbations, then integrate
                # one deterministic run for the spinup period
                # Initial condition for the nature run is the last state of the spinup run
                if verbose: print('Spinup        : ',end='')
                sp = CBL(self.cbl_settings)
                sp.maxtime = self.tspinup
                sp.initialize(1) 
                sp.run()
                initial_conditions = sp.x
                coords = sp.state_coordinates
            else:
                # No spinup, nature run gets unperturbed initial condition
                sp = CBL(self.cbl_settings)
                sp.initialize(1)
                initial_conditions = sp.x0
                coords = sp.state_coordinates

            # Nature run: initialize only
            # I.e., store "initial_conditions" in nr.x0
            self.nr = CBL(self.cbl_settings)
            self.nr.maxtime = self.trun
            self.nr.initialize(initial_conditions,coordinates=coords)

        ################################################################################################
        # Run assimilation cycle
        if not hasattr(self,'da'):
            self.da = cycle(self.cbl_settings,
                        self.nr,
                        self.nens,
                        self.FILTER,                        
                        self.trun,
                        self.tspinup_assim,
                        self.assimilation_interval,
                        self.obs_coordinates,
                        self.observations,
                        self.obs_error_sdev_assimilate,
                        self.localization_cutoff,
                        self.inflation_rtps_alpha,
                        self.simulate_underdispersiveness,
                        self.rnseed
                        )

        # If truths (from nature run) exist, add them to the da cycle structure.
        # That's necessary for the diagnostics package to work properly.
        if hasattr(self,'truths'):
            self.da.truths = self.truths

        # Compute diagnostics
        if not hasattr(self,'dg'):
            self.dg = diagnostics(self.da)

class diagnostics:
    def __init__(self,da):
        # ntim: number of verification times in the sample
        # nens: number of ensemble members
        # nvar: number of model state variables
        # nobs: number of observations
        
        # Rename variables for shorthand
        time_indices = da.analysis_times
        ntim = time_indices.size
        nobs = da.observations.shape[1]
        nens = da.backgrounds.shape[2]
        if da.obs_coordinates.ndim == 1:     # nobs or ntim*nobs
            ocoords = da.obs_coordinates
        elif da.obs_coordinates.ndim == 2:   # nobs or ntim*nobs
            ocoords = da.obs_coordinates[time_indices,:]
        xcoords = da.state_coordinates       # nvar
        o = da.observations[time_indices,:]  # ntim*nobs
        if hasattr(da,'truths'):
            t = da.truths[time_indices,:]    # ntim*nobs
            truth_exists = True
        else:
            truth_exists = False

        # Get observation-space priors from da output
        # Consider only the model equivalents beyond the assimilation spinup
        b = da.model_equivalents[time_indices,:,:] # ntim*nobs*nens

        # To get observation-space posteriors (ntim*nvar*nens),
        # apply observation operator to analyses
        # First allocate an empty array.
        a = np.zeros((ntim,nobs,nens))+np.nan

        # Consider only the analyses beyond the assimilation spinup
        analyses = da.analyses[time_indices,:,:]

        # In this case, observation coordinates do not change between analysis times
        if ocoords.ndim == 1:
            for i in range(ntim):
                for j in range(nobs):
                    for k in range(nens):
                        a[i,j,k] = observation_operator(analyses[i,:,k],xcoords,ocoords[j])

        # In this case, observation coordinates are different at each analysis time
        elif ocoords.ndim == 2:
            for i in range(ntim):
                for j in range(nobs):
                    for k in range(nens):
                        a[i,j,k] = observation_operator(analyses[i,:,k],xcoords,ocoords[i,j])

        # For the computation of time-averaged statistics, all
        # data arrays must be properly sorted in space (otherwise the elements
        # of the time average refer to different positions).
        # Re-sort the coordinates and the data arrays, in case they are not
        for k in range(ntim):
            if ocoords.ndim == 2:
                indices = np.argsort(ocoords[k,:])
            elif ocoords.ndim == 1:
                indices = np.argsort(ocoords)
            a[k,:,:] = a[k,indices,:]
            b[k,:,:] = b[k,indices,:]
            o[k,:] = o[k,indices]
            if truth_exists:
                t[k,:] = t[k,indices]

        # Compute ensemble mean and spread
        meana  = a.mean(axis=2) # ntim*nobs
        meanb  = b.mean(axis=2) # ntim*nobs
        sigmaa = a.std(axis=2)  # ntim*nobs
        sigmab = b.std(axis=2)  # ntim*nobs

        # Compute deviations
        omina = o-meana     # ntim*nobs 
        ominb = o-meanb     # ntim*nobs 
        aminb = meana-meanb # ntim*nobs 
        if truth_exists:
            tmina = t-meana # ntim*nobs 
            tminb = t-meanb # ntim*nobs 

        # Time-averaged Desroziers statistics (size nobs)
        vara_ta    = (sigmaa**2).mean(axis=0)
        varb_ta    = (sigmab**2).mean(axis=0)
        dobdob_ta  = (ominb*ominb).mean(axis=0)
        dabdob_ta  = (aminb*ominb).mean(axis=0)
        doadob_ta  = (omina*ominb).mean(axis=0)
        dabdoa_ta  = (aminb*omina).mean(axis=0)

        # Time-averaged RMSE and spread (size nobs)
        aRMSD_t = np.sqrt( (omina*omina).mean(axis=0) )
        bRMSD_t = np.sqrt(dobdob_ta)
        aSprd_t = np.sqrt(vara_ta)
        bSprd_t = np.sqrt(varb_ta)
        if truth_exists:
            aRMSE_t = np.sqrt( (tmina*tmina).mean(axis=0) )
            bRMSE_t = np.sqrt( (tminb*tminb).mean(axis=0) )

        # Space-averaged Desroziers statistics (size ntim)
        vara_xa    = (sigmaa**2).mean(axis=1)
        varb_xa    = (sigmab**2).mean(axis=1)
        dobdob_xa  = (ominb*ominb).mean(axis=1)
        dabdob_xa  = (aminb*ominb).mean(axis=1)
        doadob_xa  = (omina*ominb).mean(axis=1)
        dabdoa_xa  = (aminb*omina).mean(axis=1)

        # Space-averaged RMSE and spread (size ntim)
        aRMSD_x = np.sqrt( (omina*omina).mean(axis=1) )
        bRMSD_x = np.sqrt(dobdob_xa)
        aSprd_x = np.sqrt(vara_xa)
        bSprd_x = np.sqrt(varb_xa)
        if truth_exists:
            aRMSE_x = np.sqrt( (tmina*tmina).mean(axis=1) )
            bRMSE_x = np.sqrt( (tminb*tminb).mean(axis=1) )

        # Space/time-averaged Desroziers statistics (scalar)
        vara_xta   = (sigmaa**2).mean()
        varb_xta   = (sigmab**2).mean()
        dobdob_xta = (ominb*ominb).mean()
        dabdob_xta = (aminb*ominb).mean()
        doadob_xta = (omina*ominb).mean()
        dabdoa_xta = (aminb*omina).mean()

        # Space/time-averaged RMSE and spread (scalar)
        aRMSD_xt = np.sqrt( (omina*omina).mean() )
        bRMSD_xt = np.sqrt(dobdob_xta)
        aSprd_xt = np.sqrt(vara_xta)
        bSprd_xt = np.sqrt(varb_xta)
        if truth_exists:
            aRMSE_xt = np.sqrt( (tmina*tmina).mean() )
            bRMSE_xt = np.sqrt( (tminb*tminb).mean() )

        # Compute histograms of ensemble mean departures.
        # Histograms are meaningless if you only have 1 observation.
        if o.size > 1:
            nbins = 100
            #
            edges = np.linspace(-omina.std()*4,omina.std()*4,nbins+1)
            hist_omina_global   = np.histogram(omina,bins=edges)[0]/omina.size/(edges[1]-edges[0])
            hist_omina_x = 0.5*(edges[1:]+edges[:-1])
            hist_omina_local = []
            for i in range(nobs):
                hist_omina_local.append(np.histogram(omina[:,i],bins=edges)[0]/omina[:,i].size/(edges[1]-edges[0]))
            #
            edges = np.linspace(-ominb.std()*4,ominb.std()*4,nbins+1)
            hist_ominb_global   = np.histogram(ominb,bins=edges)[0]/ominb.size/(edges[1]-edges[0])
            hist_ominb_x = 0.5*(edges[1:]+edges[:-1])
            hist_ominb_local = []
            for i in range(nobs):
                hist_ominb_local.append(np.histogram(ominb[:,i],bins=edges)[0]/ominb[:,i].size/(edges[1]-edges[0]))
            #
            edges = np.linspace(-aminb.std()*4,aminb.std()*4,nbins+1)
            hist_aminb_global   = np.histogram(aminb,bins=edges)[0]/aminb.size/(edges[1]-edges[0])
            hist_aminb_x = 0.5*(edges[1:]+edges[:-1])
            hist_aminb_local = []
            for i in range(nobs):
                hist_aminb_local.append(np.histogram(aminb[:,i],bins=edges)[0]/aminb[:,i].size/(edges[1]-edges[0]))

        # All these with dimensions ntim*nobs
        self.meana  = meana
        self.meanb  = meanb
        self.sigmaa = sigmaa
        self.sigmab = sigmab
        self.omina  = omina
        self.ominb  = ominb
        self.aminb  = aminb

        # All these with dimensions ntim
        self.aRMSD_x = aRMSD_x
        self.aSprd_x = aSprd_x
        self.bRMSD_x = bRMSD_x
        self.bSprd_x = bSprd_x
        self.dobdob_xa = dobdob_xa
        self.dabdob_xa = dabdob_xa
        self.doadob_xa = doadob_xa
        self.dabdoa_xa = dabdoa_xa
        self.vara_xa = vara_xa
        self.varb_xa = varb_xa

        # All these with dimensions nobs
        self.aRMSD_t = aRMSD_t
        self.aSprd_t = aSprd_t
        self.bRMSD_t = bRMSD_t
        self.bSprd_t = bSprd_t
        self.dobdob_ta = dobdob_ta
        self.dabdob_ta = dabdob_ta
        self.doadob_ta = doadob_ta
        self.dabdoa_ta = dabdoa_ta
        self.vara_ta = vara_ta
        self.varb_ta = varb_ta
        self.varo    = da.obs_error_sdev_assimilate**2
        self.sigma_ominb = ominb.std(axis=0)
        self.mean_ominb  = ominb.mean(axis=0)
        self.sigma_omina = omina.std(axis=0)
        self.mean_omina  = omina.mean(axis=0)

        # All these are scalars
        self.aRMSD_xt = aRMSD_xt
        self.aSprd_xt = aSprd_xt
        self.bRMSD_xt = bRMSD_xt
        self.bSprd_xt = bSprd_xt
        self.dobdob_xta = dobdob_xta
        self.dabdob_xta = dabdob_xta
        self.doadob_xta = doadob_xta
        self.dabdoa_xta = dabdoa_xta
        self.sigma_ominb_global = ominb.std()
        self.mean_ominb_global  = ominb.mean()
        self.sigma_omina_global = omina.std()
        self.mean_omina_global  = omina.mean()

        # Additional stuff if truth exists
        if truth_exists:
            self.aRMSE_x = aRMSE_x   # ntim
            self.bRMSE_x = bRMSE_x   # ntim
            self.aRMSE_t = aRMSE_t   # nobs
            self.bRMSE_t = bRMSE_t   # nobs
            self.aRMSE_xt = aRMSE_xt # scalar
            self.bRMSE_xt = bRMSE_xt # scalar

        # Store histograms, if they were computed
        if o.size > 1:
            # All these with dimensions nobs * nbins
            self.hist_omina_local = hist_omina_local
            self.hist_ominb_local = hist_ominb_local
            self.hist_aminb_local = hist_aminb_local

            # All these with dimensions nbins
            self.hist_omina_global = hist_omina_global
            self.hist_omina_x      = hist_omina_x
            self.hist_ominb_global = hist_ominb_global
            self.hist_ominb_x      = hist_ominb_x
            self.hist_aminb_global = hist_aminb_global
            self.hist_aminb_x      = hist_aminb_x