#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import numpy as np
from matplotlib import pyplot as p
import matplotlib.colors as colors
from ENDA import compute_distances, gaspari_cohn

is_sorted = lambda a: np.all(a[:-1] <= a[1:])

def truncate_colormap(cmap, minval=0.0, maxval=1.0, n=100):
    #https://stackoverflow.com/questions/18926031/how-to-extract-a-subset-of-a-colormap-as-a-new-colormap-in-matplotlib
    new_cmap = colors.LinearSegmentedColormap.from_list(
        'trunc({n},{a:.2f},{b:.2f})'.format(n=cmap.name, a=minval, b=maxval),
        cmap(np.linspace(minval, maxval, n)))
    return new_cmap

def plot_CBL_assimilation(exp,figtitle,which_cycle=0,ax=None):

    naturerun_color = 'black'
    ensmembers_color = 'lightskyblue'
    ensmean_color = 'royalblue'
    
    def ensplot(ax,x,x1,x2,z,colors=[ensmean_color,ensmembers_color],label=None):
        nens = x.shape[1]
        ax.plot(x[x1:x2].mean(axis=1),z,color=colors[0],zorder=5,label=label)
        for i in range(nens): ax.plot(x[x1:x2,i],z,color=colors[1],alpha=0.2,zorder=1)

    # Shorthand
    o = exp.observations[which_cycle]
    b = exp.da.backgrounds[which_cycle,:,:]
    a = exp.da.analyses[which_cycle,:,:]
    if exp.obs_coordinates.ndim == 1:
        ocoords = exp.obs_coordinates
    elif exp.obs_coordinates.ndim == 2:
        ocoords = exp.obs_coordinates[which_cycle]
    nz = exp.nr.nz
    zt = exp.nr.zt 

    if exp.nr.is_bwind:
        ii=3
    else:
        ii=1

    zmax = 2000

    if ax is None:
        fig = p.figure(151)
        fig.set_size_inches(9,3*ii)
        #
        # Potential temperature
        #
        ax1 = fig.add_subplot(ii,4,1)
        ax1.scatter(o,ocoords,marker="o",s=20,color=naturerun_color,zorder=10)
        ax1.set_title(r'$\theta$ (K), nature run')
        ax1.set_ylabel(r'$z$ (m)')
        ax1.set_ylim([0,zmax])
        #
        ax2 = fig.add_subplot(ii,4,2,sharey=ax1)
        ensplot(ax2,b,0,nz,zt)
        ax2.scatter(o,ocoords,marker="o",s=20,color=naturerun_color,zorder=10)
        ax2.set_title(r'$\theta$ (K), prior')
        #
        ax3 = fig.add_subplot(ii,4,3,sharey=ax1)
        ensplot(ax3,a,0,nz,zt)
        ax3.scatter(o,ocoords,marker="o",s=20,color=naturerun_color,zorder=10)
        ax3.set_title(r'$\theta$ (K), posterior')
        #
        ax4 = fig.add_subplot(ii,4,4,sharey=ax1)
        ax4.plot(a[:nz].mean(axis=1)-b[:nz].mean(axis=1),zt,color=ensmean_color)
        ax4.set_title(r'$\theta$ (K), mean increment',zorder=10)
        #
        # Fix axis labels
        #
        for ax in [ax2,ax3,ax4]:
            p.setp(ax.get_yticklabels(), visible=False)
        #
        if exp.nr.is_bwind:
            #
            # u wind component
            #
            ax5 = fig.add_subplot(ii,4,5)
            ax5.set_title(r'$u$ (m/s), nature run')
            ax5.set_ylabel(r'$z$ (m)')
            ax5.set_ylim([0,zmax])
            #
            ax6 = fig.add_subplot(ii,4,6,sharey=ax1)
            ensplot(ax6,b,nz,nz*2,zt)
            ax6.set_title(r'$u$ (m/s), prior')
            #
            ax7 = fig.add_subplot(ii,4,7,sharey=ax1)
            ensplot(ax7,a,nz,nz*2,zt)
            ax7.set_title(r'$u$ (m/s), posterior')
            #
            ax8 = fig.add_subplot(ii,4,8,sharey=ax1)
            ax8.plot(a[nz:nz*2].mean(axis=1)-b[nz:nz*2].mean(axis=1),zt,color=ensmean_color)
            ax8.set_title(r'$u$ (m/s), mean increment',zorder=10)
            #
            # v wind component
            #
            ax9 = fig.add_subplot(ii,4,9)
            ax9.set_title(r'$v$ (m/s), nature run')
            ax9.set_ylabel(r'$z$ (m)')
            ax9.set_ylim([0,zmax])
            #
            axa = fig.add_subplot(ii,4,10,sharey=ax1)
            ensplot(axa,b,nz*2,nz*3,zt)
            axa.set_title(r'$v$ (m/s), prior')
            #
            axb = fig.add_subplot(ii,4,11,sharey=ax1)
            ensplot(axb,a,nz*2,nz*3,zt)
            axb.set_title(r'$v$ (m/s), posterior')
            #
            axc = fig.add_subplot(ii,4,12,sharey=ax1)
            axc.plot(a[nz*2:nz*3].mean(axis=1)-b[nz*2:nz*3].mean(axis=1),zt,color=ensmean_color)
            axc.set_title(r'$v$ (m/s), mean increment',zorder=10)
            #
            # Fix axis labels
            #
            for ax in [ax6,ax7,ax8,axa,axb,axc]:
                p.setp(ax.get_yticklabels(), visible=False)
        #
        fig.savefig(figtitle,format='png',dpi=300)
        p.close(fig)
    
    else:

        ensplot(ax,b,0,nz,zt,colors = ['red','salmon'],label='prior')
        ensplot(ax,a,0,nz,zt,colors = ['royalblue','lightskyblue'],label='posterior')
        ax.scatter(o,ocoords,marker="o",s=20,color=naturerun_color,zorder=10,label='observations')
        #ax.set_title(r'$\theta$ (K), posterior')
        return ax

def plot_CBL_PE(exp,figtitle,parameter_id=0,show_spread=False,show_true_parameter=False,show_transformed_parameter=False,plot_parameter_histogram = False,ax=None):

    if exp.nr.do_parameter_estimation:

        spread_color = 'lightskyblue'
        mean_color = 'royalblue'
        mean_color_2 = 'orange'

        pt = exp.nr.parameter_transform
        if parameter_id==0:
            true_value = exp.nr.pfac
            true_value_transformed = pt[0](exp.nr.pfac,kind='dir')

        parameter_number = exp.nr.parameter_number
        par_tran = exp.da.backgrounds[:,-parameter_number:,:][:,parameter_id,:]
        par_phys = pt[parameter_id](par_tran,kind='inv')
        initpar_phys = pt[parameter_id](exp.da.initial_perturbed_parameters,kind='inv')
        initpar_tran = exp.da.initial_perturbed_parameters
        t = exp.da.time/3600.

        if plot_parameter_histogram:
            p10beg_tran = np.percentile(par_tran[0,:],  10)
            p90beg_tran = np.percentile(par_tran[0,:],  90)
            p10end_tran = np.percentile(par_tran[-1,:], 10)
            p90end_tran = np.percentile(par_tran[-1,:], 90)
            p10beg_phys = np.percentile(par_phys[0,:],  10)
            p90beg_phys = np.percentile(par_phys[0,:],  90)
            p10end_phys = np.percentile(par_phys[-1,:], 10)
            p90end_phys = np.percentile(par_phys[-1,:], 90)
            #
            fig = p.figure(151)
            fig.set_size_inches(6,3)
            #
            ax1 = fig.add_subplot(1,2,1)
            ax1.hist(par_phys[0,:], bins=np.linspace(0,5,51), density=True, histtype='step', color='blue', label='initial')
            ax1.axvspan(p10beg_phys,p90beg_phys,color='blue',lw=0,alpha=0.1)
            ax1.hist(par_phys[-1,:], bins=np.linspace(0,5,51), density=True, histtype='step', color='red', label='final')
            ax1.axvspan(p10end_phys,p90end_phys,color='red',lw=0,alpha=0.1)
            ax1.axvline(x=exp.nr.parameter_true,color='black',dashes=[3,1])
            ax1.set_xlabel(r'$p$')
            ax1.set_ylabel('probability density')
            #
            ax2 = fig.add_subplot(1,2,2)
            ax2.hist(par_tran[0,:], bins=np.linspace(-10,10,51), density=True, histtype='step', color='blue', label='initial')
            ax2.axvspan(p10beg_tran,p90beg_tran,color='blue',lw=0,alpha=0.1)
            ax2.hist(par_tran[-1,:], bins=np.linspace(-10,10,51), density=True, histtype='step', color='red', label='final')
            ax2.axvspan(p10end_tran,p90end_tran,color='red',lw=0,alpha=0.1)
            ax2.axvline(x=pt[parameter_id](exp.nr.parameter_true,kind='dir'),color='black',dashes=[3,1])
            ax2.set_xlabel(r"$p\prime$")
            ax2.set_ylabel('probability density')
            #
            fig.tight_layout()
            fig.savefig('p_histogram_%s.png'%exp.label,format='png',dpi=300)
            p.close(fig)

        if ax is None:
            printfigure = True
        else:
            printfigure = False

        if printfigure:
            if show_spread:
                ii=2
            else:
                ii=1
            fig = p.figure(151)
            fig.set_size_inches(4,3*ii)
            ax1 = fig.add_subplot(ii,1,1)
        else:
            ax1 = ax
        
        if show_transformed_parameter:
            par =  par_tran
            initpar = initpar_tran
        else:
            par = par_phys
            initpar = initpar_phys

        # Initial parameter distribution
        ax1.step([0,t[0]],np.median(initpar,axis=1)*np.array([1,1]),color=mean_color,label=r'ensemble median')
        ax1.step([0,t[0]],np.mean(initpar,axis=1)*np.array([1,1]),color=mean_color_2,label=r'ensemble mean, $\mu$')
        bmin = np.percentile(initpar, 10)
        bmax = np.percentile(initpar, 90)
        ax1.fill_between([0,t[0]],\
                    y1=[bmin,bmin],\
                    y2=[bmax,bmax],\
                    color=spread_color,
                    edgecolor=None,
                    alpha=0.3,
                    label=r'10-90 percentile range')
        # Later times
        ax1.step(t,np.median(par,axis=1),color=mean_color)
        ax1.step(t,np.mean(par,axis=1),color=mean_color_2)
        for i in range(t.size-1):
            bmin = np.percentile(par, 10, axis=1)[i+1]
            bmax = np.percentile(par, 90, axis=1)[i+1]
            ax1.fill_between(t[i:i+2],\
                        y1=[bmin,bmin],\
                        y2=[bmax,bmax],\
                        color=spread_color,
                        edgecolor=None,
                        alpha=0.3)
        if show_true_parameter:
            if show_transformed_parameter:
                ax1.axhline(y=true_value_transformed,linestyle='--',color='black')
            else:
                ax1.axhline(y=true_value,linestyle='--',color='black')
        ax1.set_xlim([0,t.max()])

        if (not printfigure) and show_spread:
            bmin = (np.mean(initpar)-np.std(initpar))
            bmax = (np.mean(initpar)+np.std(initpar))
            ax1.fill_between([0,t[0]],\
                        y1=[bmin,bmin],\
                        y2=[bmax,bmax],\
                        color='lightgrey',
                        edgecolor=None,
                        alpha=0.3,
                        zorder=10,
                        label=r'$\mu\pm\sigma$')
            for i in range(t.size-1):
                bmin = (np.mean(par,axis=1)-np.std(par,axis=1))[i+1]
                bmax = (np.mean(par,axis=1)+np.std(par,axis=1))[i+1]
                ax1.fill_between(t[i:i+2],\
                            y1=[bmin,bmin],\
                            y2=[bmax,bmax],\
                            color='lightgrey',
                            edgecolor=None,
                            alpha=0.3,
                            zorder=10)
        ax1.legend(loc=1,frameon=False,fontsize=6)

        if printfigure and show_spread:
            ax2 = fig.add_subplot(ii,1,2)
            ax2.plot(t,par_phys.std(axis=1),color=mean_color,label='physical')
            ax2.plot(t,par_tran.std(axis=1),color=mean_color_2,label='transformed', linestyle='--')
            ax2.legend()
            ax2.set_title('Parameter spread')

        if printfigure:
            fig.tight_layout()
            fig.savefig(figtitle,format='png',dpi=300)
            p.close(fig)
        else:
            return ax1

def plot_CBL_identifiability(cbl,sigma_o,ax=None):

    cmap = p.get_cmap('RdBu_r')
    new_cmap = truncate_colormap(cmap, 0.5, 1.0)
    zmax = 2000
    ncont = 13

    # cbl must be an "experiment" instance
    # Read relevant dimensions
    zt = cbl.da.zt
    nz = cbl.da.nz
    nens = cbl.da.nens
    nobs = cbl.da.observations.shape[1]
    npar = cbl.nr.parameter_number
    ntimes = cbl.da.ncycles
    times = cbl.da.time

    # Expand times array to 2D (matching the number of observations)
    times = times[:,None]+np.zeros((ntimes,nobs))

    # Read model equivalents and state vector elements
    #theta = cbl.da.model_equivalents  # ntim*nobs*nens
    pfac = cbl.da.backgrounds[:,-npar,:] # ntim*nx*nens
    pfac = cbl.nr.parameter_transform[0](pfac,kind='inv')

    # Read observation coordinates
    if cbl.da.obs_coordinates.ndim == 1:
        zz = cbl.da.obs_coordinates
        zz = zz[None,:]+np.zeros(times.shape)
    elif cbl.da.obs_coordinates.ndim == 2:
        zz = cbl.da.obs_coordinates

    # At each time, check if the coordinate array is sorted
    # if not, resort it.
    theta      = np.zeros((ntimes,nobs,nens))+np.nan
    zobs       = np.zeros((ntimes,nobs))+np.nan
    innov      = np.zeros((ntimes,nobs))+np.nan
    cov_yy     = np.zeros((ntimes,nobs))+np.nan
    cov_pp     = np.zeros((ntimes,nobs,npar))+np.nan
    cov_py     = np.zeros((ntimes,nobs,npar))+np.nan
    increments = np.zeros((ntimes,nobs,npar))+np.nan

    for k in range(ntimes):
        if is_sorted(zz[k,:]):
            indices = np.arange(nobs)
        else:
            indices = np.argsort(zz[k,:])
        zobs[k,:] = zz[k,indices]
        theta[k,:,:] = cbl.da.model_equivalents[k,indices,:]
        innov[k,:] = cbl.da.innov[k,indices] # ntim*nobs
        cov_yy[k,:] = cbl.da.cov_yy[k,indices] # ntim*nobs
        cov_pp[k,:,:] = cbl.da.cov_pp[k,indices,:] # ntim*nobs(*npar)
        cov_py[k,:,:] = cbl.da.cov_py[k,indices,:] # ntim*nobs(*npar)
        increments[k,:,:] = cbl.da.increments[k,indices,:] # ntim*nobs(*npar)

    if hasattr(cbl.da,'cov_py'):
        sigma2_yb = cov_yy # ntim*nobs
        sigma2_p = np.squeeze(cov_pp) # ntim*nobs(*npar)
        covariance = np.squeeze(cov_py) # ntim*nobs(*npar)
        correlation = covariance/(sigma2_yb*sigma2_p)**0.5
        deltap = np.squeeze(increments) # ntim*nobs(*npar)

    # Shorthand for plots
    A = correlation.T
    B = ( innov*(sigma2_p*sigma2_yb)**0.5/(sigma2_yb+sigma_o**2) ).T
    C = deltap.T
    D = ( (sigma2_p/sigma2_yb)**0.5 ).T
    E = ( innov*sigma2_yb/(sigma2_yb+sigma_o**2) ).T
    F = innov.T

    # For plotting
    contours = np.linspace(cbl.nr.theta_0,cbl.nr.theta_0+cbl.nr.gamma*zmax,ncont)

    # Before plotting, transpose the 2D arrays of times and coordinates
    times = times.T
    zobs = zobs.T

    # Make plots
    if ax is None:
        pass

    elif type(ax) is p.Axes:

        c0 = ax.pcolormesh(times/3600.,zobs,A,vmin=-1,vmax=1,cmap='RdBu_r')
        ax.contour(times/3600.,zobs,theta.mean(axis=2).T,
                    contours,
                    colors='black',
                    linestyles='--',
                    linewidths=0.75)
        ax.set_ylim([0,zmax])

        return ax,c0

    elif type(ax) is list and len(ax)==2:

        #c0 = ax[0].pcolormesh(times/3600.,zobs,D,cmap=new_cmap) # For old version of Fig 8
        c0 = ax[0].pcolormesh(times/3600.,zobs,B,norm=colors.CenteredNorm(),cmap='RdBu_r')
        ax[0].contour(times/3600.,zobs,theta.mean(axis=2).T,
                    contours,
                    colors='black',
                    linestyles='--',
                    linewidths=0.75)
        ax[0].set_ylim([0,zmax])

        #c1 = ax[1].pcolormesh(times/3600.,zobs,E,norm=colors.CenteredNorm(),cmap='RdBu_r') # For old version of Fig 8
        c1 = ax[1].pcolormesh(times/3600.,zobs,C,norm=colors.CenteredNorm(),cmap='RdBu_r')
        ax[1].contour(times/3600.,zobs,theta.mean(axis=2).T,
                    contours,
                    colors='black',
                    linestyles='--',
                    linewidths=0.75)
        ax[1].set_ylim([0,zmax])

        return ax,c0,c1

    elif type(ax) is list and len(ax)==3:

        c0 = ax[0].pcolormesh(times/3600.,zobs,A,vmin=-1,vmax=1,cmap='RdBu_r')
        ax[0].contour(times/3600.,zobs,theta.mean(axis=2).T,
                    contours,
                    colors='black',
                    linestyles='--',
                    linewidths=0.75)
        ax[0].set_ylim([0,zmax])

        c1 = ax[1].pcolormesh(times/3600.,zobs,B,norm=colors.CenteredNorm(),cmap='RdBu_r')
        ax[1].contour(times/3600.,zobs,theta.mean(axis=2).T,
                    contours,
                    colors='black',
                    linestyles='--',
                    linewidths=0.75)
        ax[1].set_ylim([0,zmax])

        c2 = ax[2].pcolormesh(times/3600.,zobs,C,norm=colors.CenteredNorm(),cmap='RdBu_r')
        ax[2].contour(times/3600.,zobs,theta.mean(axis=2).T,
                    contours,
                    colors='black',
                    linestyles='--',
                    linewidths=0.75)
        ax[2].set_ylim([0,zmax])

    make_parameter_state_scatterplots = False
    if make_parameter_state_scatterplots:
        fig, [[ax1, ax2],[ax3, ax4]] = p.subplots(2,2,constrained_layout=True)
        fig.set_size_inches(4,4)
        ax3.scatter(theta[:,0,0] , pfac) #ll
        ax4.scatter(theta[:,0,-1], pfac) #lr
        ax1.scatter(theta[:,-1,0], pfac) #ul
        ax2.scatter(theta[:,-1,-1],pfac) #ur
        fig.savefig('p_scatter_%s.png'%cbl.label,format='png',dpi=300)
        p.close(fig)

    return ax,c0,c1,c2

def plot_p(p_factors,theta_profiles,zt,figtitle,ax=None):

    zoverh = np.linspace(0,1,101)

    if ax is None:
        fig = p.figure(151)
        fig.set_size_inches(6,3)
        #
        ax1 = fig.add_subplot(1,2,1)
        for pfac in p_factors:
            Koverkws = zoverh*(1-zoverh)**pfac
            ax1.plot(Koverkws,zoverh,label='$p=%4.1f$'%pfac)
        ax1.set_xlabel('$K/(\kappa w_s z_i)$')
        ax1.set_ylabel('$z/z_i$')
        #
        ax2 = fig.add_subplot(1,2,2)
        for i in range(len(p_factors)):
            ax2.plot(theta_profiles[i],zt,label='$p=%4.1f$'%p_factors[i])
        ax2.set_xlabel(r'$\theta$ (K)')
        ax2.set_ylabel(r'$z$ (m)')
        ax2.set_ylim([0,1500])
        ax2.set_xlim([291,297])
        ax2.legend(loc=4)
        #
        fig.tight_layout()
        fig.savefig(figtitle,format='png',dpi=300)
        p.close(fig)  

    else:
        for i in range(len(p_factors)):
            ax.plot(theta_profiles[i],zt,label='$p=%4.1f$'%p_factors[i])
        return ax

def plot_free_ensemble(cbl,show='spread',ax=None,cmap='viridis'):

    # Read parameters
    param = cbl.initial_perturbed_parameters

    # Read relevant dimensions
    times = cbl.history['0000']['time']
    zt = cbl.zt
    nz = cbl.nz
    ntimes = times.size
    nens = cbl.nens

    # Reconstruct_history
    theta = np.zeros((nens,nz,ntimes))+np.nan
    for k in range(nens):
        theta[k,:,:]= cbl.history['%04u'%k]['theta']

    # Compute spread
    if show=='spread':
        bkgd = theta.std(axis=0)
        contour_color = 'white'
    elif show=='mean':
        bkgd = theta.mean(axis=0)
        contour_color = 'white'
    elif show=='cov':
        cov = np.zeros((nz,ntimes))+np.nan
        for ii in range(nz):
            for jj in range(ntimes):
                cov[ii,jj] = np.cov(param,theta[:,ii,jj])[0,1]
        bkgd = cov
        contour_color = 'black'
    elif show=='corr':
        corr = np.zeros((nz,ntimes))+np.nan
        for ii in range(nz):
            for jj in range(ntimes):
                corr[ii,jj] = np.corrcoef(param,theta[:,ii,jj])[0,1]
        bkgd = corr
        contour_color = 'black'

    # Plot
    zmax = 2000
    ncont = 13

    if ax is not None:
        # Make plots
        if show=='cov' or show=='corr':
            c = ax.pcolormesh(times/3600.,zt,bkgd,cmap=cmap,norm=colors.CenteredNorm())
        else:
            c = ax.pcolormesh(times/3600.,zt,bkgd,cmap=cmap)
        ax.contour(times/3600.,zt,theta.mean(axis=0),
                    np.linspace(cbl.theta_0,cbl.theta_0+cbl.gamma*zmax,ncont),
                    colors=contour_color,
                    linestyles='--',
                    linewidths=0.75)
        ax.set_ylim([0,zmax])
        return ax,c

def plot_diagnostics(exp,show='errors',axes=None,label='',linecolor='blue',zmax=None):

    # See if the experiment is an OSSE
    if (not hasattr(exp.da,'truths')):
        truth_exists = False
    else:
        truth_exists = True

    # Get the data
    sigma_o = exp.dg.varo**0.5
    sigma_d = exp.dg.sigma_ominb
    armsd = exp.dg.aRMSD_t
    brmsd = exp.dg.bRMSD_t
    #bsprd = exp.dg.bSprd_t
    bsprd = exp.dg.varb_ta**0.5
    if truth_exists:
        armse = exp.dg.aRMSE_t
        brmse = exp.dg.bRMSE_t

    # Read observation coordinates
    # In the 2D case, it is assumed that the set of 
    # observation coordinates does not change
    # from time to time. Anyway, coordinates need
    # to be sorted. Diagnostic quantities are 
    # already computed in a way that takes the correct
    # spatial sorting into account.

    if exp.da.obs_coordinates.ndim == 1:
        z = np.sort(exp.da.obs_coordinates)
    elif exp.da.obs_coordinates.ndim == 2:
        z = np.sort(exp.da.obs_coordinates[0,:])

    if zmax is None:
        idx = np.arange(z.size)
    else:
        idx = (z<zmax)

    if axes is not None and len(axes)==4:

        # Plot errors (wrt to true state) only if true state exists
        if show=='errors' and truth_exists:

            axes[0].plot(armse[idx],z[idx],label=label,color=linecolor)#,color=linecolors[i])
            axes[1].plot(brmse[idx],z[idx],label=label,color=linecolor)
            axes[2].plot(brmse[idx]-armse[idx],z[idx],label=label,color=linecolor)
            axes[3].plot(np.sqrt(bsprd[idx]**2+sigma_o[idx]**2)/brmse[idx],z[idx],label=label,color=linecolor)

        elif show=='deviations':

            # Compute time-average mean bias of background and analyses
            mean_bias_a_i1 = exp.dg.omina.mean(axis=0)
            mean_bias_b_i1 = exp.dg.ominb.mean(axis=0)
            # Plots
            axes[0].plot(armsd[idx],z[idx],label=label,color=linecolor)
            axes[1].plot(brmsd[idx],z[idx],label=label,color=linecolor)
            axes[2].plot(brmsd[idx]-armsd[idx],z[idx],label=label,color=linecolor)
            axes[3].plot(np.sqrt(bsprd[idx]**2+sigma_o[idx]**2)/brmsd[idx],z,label=label,color=linecolor)

    return axes

def plot_linearity(cbl,ax=None):
    
    # Read parameters
    param = np.squeeze(cbl.parameter_transform[0](cbl.initial_perturbed_parameters,kind='inv'))

    # Get the indices that sort parameters in ascending order
    indices = np.argsort(param)

    # Read relevant dimensions
    times = cbl.history['0000']['time']
    zt = cbl.zt
    nz = cbl.nz
    ntimes = times.size
    nens = cbl.nens

    # Reconstruct_history
    theta = np.zeros((nens,nz,ntimes))+np.nan
    for k in range(nens):
        theta[k,:,:]= cbl.history['%04u'%k]['theta']

    # Make plot
    if ax is not None:
        tt = -1 # pick the last time in the history
        zz = [0,10,20,30] # pick a few heights
        hh = [int(cbl.dz/2)+i*cbl.dz for i in zz]
        labels = [r'$z=%s$ m'%i for i in hh]
        for z,label in zip(zz,labels):
            #ax.scatter(param[indices],theta[indices,z,tt],s=5,label=label)
            ax.plot(param[indices],theta[indices,z,tt],label=label)
        ax.legend(loc=3,frameon=False,fontsize=6,facecolor='lightgrey')

    return ax

def plot_B_matrix(exp,timeid=-1,localize=True):

    # Read the background state for the requested analysis time
    xx = exp.da.backgrounds[timeid]
    nvar,nens = xx.shape

    # Analysis time
    tt = exp.da.time[timeid]

    # Compute deviations from ensemble mean
    X = xx.T-np.tile(xx.mean(axis=1),(nens,1))

    # Standardize deviations from ensemble mean
    X = X/np.tile(xx.std(axis=1),(nens,1))

    # Compute covariance matrix
    Pb = np.matmul(X.T,X)/(nens-1)
    v = np.abs(Pb[:-1,:-1]).max()

    # Compute localized covariance matrix
    xcoords = exp.nr.zt
    dist = compute_distances(xcoords,xcoords)
    loc = np.ones((nvar,nvar))
    loc[:(nvar-1),:(nvar-1)] = gaspari_cohn(dist,cutoff=800)

    # Plot
    fig, [sp1,sp2] = p.subplots(1,2,constrained_layout=True)
    fig.set_size_inches(6,3)
    #
    c1 = sp1.pcolormesh(range(nvar),range(nvar),Pb,vmin=-v,vmax=v,cmap='RdBu_r',shading='nearest')
    sp1.set_aspect('equal','box')
    sp1.set_xlim(-0.5,nvar-0.5)
    sp1.set_ylim(-0.5,nvar-0.5)
    sp1.set_title('a) $P^b$, $t=%4.2f$ h'%(tt/3600))
    #
    c2 = sp2.pcolormesh(range(nvar),range(nvar),loc*Pb,vmin=-v,vmax=v,cmap='RdBu_r',shading='nearest')
    sp2.set_aspect('equal','box')
    sp2.set_xlim(-0.5,nvar-0.5)
    sp2.set_ylim(-0.5,nvar-0.5)
    sp2.set_title('b) Localized $P^b$, $t=%4.2f$ h'%(tt/3600))
    #
    p.colorbar(c1,orientation='horizontal')
    p.colorbar(c2,orientation='horizontal')
    fig.savefig('Covariance_matrix_%05u_s.png'%tt,format='png',dpi=150)
    p.close(fig)

def plot_desroziers(dgs,filename='desroziers_diagnostics.png',labels=None,logscale=False,ax=None):
    from matplotlib.ticker import MaxNLocator

    newfig = False    
    if ax is None:
        newfig = True
        fig = p.figure(0)
        fig.set_size_inches(4,3)
        ax = fig.add_subplot(1,1,1)

    # Check if all experiments have the same obs error variance
    all_varo_equal = False
    varo =[]
    for dg in dgs:
        varo.append(dg.varo)
    varo = np.array(varo)
    if np.all(np.isclose(varo, varo[0,0])):
        varo = varo[0,0]
        all_varo_equal = True


    colors = p.rcParams['axes.prop_cycle'].by_key()['color'][:len(dgs)]
    for i,dg,color in zip( range(len(dgs)) , dgs , colors):
        # Select time interval to consider
        idx = np.arange(0,dg.ominb.shape[0]) # whole run
        #idx = np.arange(0,24) # first two hours
        if labels is not None:
            # Two formulations of the same thing; they give the same numbers!
            #ax1.scatter( dg.dobdob_ta-dg.mean_ominb**2 , dg.varo+dg.varb_ta , s=5, label = labels[i])
            ax.scatter( np.mean(dg.ominb[idx,:].std(axis=0)**2) , np.mean(varo+(dg.sigmab[idx,:]**2).mean(axis=0)) , s=13, label = labels[i],zorder=10, c=color, edgecolors='black')
            ax.scatter( dg.ominb[idx,:].std(axis=0)**2 , varo+(dg.sigmab[idx,:]**2).mean(axis=0) , s=10, alpha = 0.3, c=color, linewidths=0)
        else:
            # Two formulations of the same thing; they give the same numbers!
            #ax1.scatter( dg.dobdob_ta-dg.mean_ominb**2 , dg.varo+dg.varb_ta , s=5)
            ax.scatter( np.mean(dg.ominb[idx,:].std(axis=0)**2) , np.mean(varo+(dg.sigmab[idx,:]**2).mean(axis=0)) , s=13, zorder = 10)
            ax.scatter( dg.ominb[idx,:].std(axis=0)**2 , varo+(dg.sigmab[idx,:]**2).mean(axis=0) , s=10, alpha = 0.3) # 
        if all_varo_equal:
            ax.axhline( y=varo, dashes=[3,1], color='lightgrey')
        #ax.set_title("Consistency of innovations",fontsize=8)
        ax.set_xlabel(r'${\langle}\sigma_d^2{\rangle}$ (K)') # '${\langle}d_{ob}^2{\rangle}_t$'
        ax.set_ylabel(r'$\sigma_{y^o}^2+{\langle}\sigma_{y^b}^2{\rangle}$ (K)')
    
    if labels is not None:
        ax.legend(frameon=False, fontsize=6, ncol=2)

    if logscale:
        ax.axline((dgs[0].varo[0],dgs[0].varo[0]), slope=1, dashes=[3,2], color='k')
        ax.set_xscale("log")
        ax.set_yscale("log")
    else:
        ax.axline((0,0), slope=1, dashes=[3,2], color='k')
        ax.axis("square")
        ax.xaxis.set_major_locator(MaxNLocator(5))
        ax.yaxis.set_major_locator(MaxNLocator(5))

    if newfig:
        fig.tight_layout()
        fig.savefig(filename,format='png',dpi=150)
        p.close(fig)

    return ax

def plot_overview(exp,label,filename,show_spread=False,show_transformed_parameter=False,show_true_parameter=True):

    maxtime = exp.nr.maxtime
    fig, [[ax0, ax1], [ax2, ax3]] = p.subplots(2,2,constrained_layout=True)
    fig.set_size_inches(6,6)
    #
    [ax0,ax1,ax2],c0,c1,c2 = plot_CBL_identifiability(exp,exp.obs_error_sdev_assimilate,ax=[ax0,ax1,ax2])
    ax0.set_title(r'a) Exp. %s, $\rho(p\prime\prime,y_b}$)'%label)
    ax0.set_xlabel('Time (h)')
    ax0.set_ylabel('Height (m)')
    ax1.set_title(r'b) Exp. %s, $\delta\overline{y}\cdot(\sigma_{p\prime\prime}/\sigma_{y^b})$'%label)
    ax1.set_xlabel('Time (h)')
    ax1.set_ylabel('Height (m)')
    ax2.set_title(r'c) Exp. %s, $\delta\overline{p}\prime\prime$'%label)
    ax2.set_xlabel('Time (h)')
    ax2.set_ylabel('Height (m)')
    ax0.set_xlim([0,maxtime/3600])
    ax1.set_xlim([0,maxtime/3600])
    ax2.set_xlim([0,maxtime/3600])
    ax0.set_xticks(np.arange((maxtime+1)/3600))
    ax1.set_xticks(np.arange((maxtime+1)/3600))
    ax2.set_xticks(np.arange((maxtime+1)/3600))
    ax3 = plot_CBL_PE(exp,None,ax=ax3,show_spread=show_spread,show_transformed_parameter=show_transformed_parameter,show_true_parameter=show_true_parameter)
    if ax3 is None:
        pass
    else:
        if show_transformed_parameter:
            ax3.set_title(r'd) Exp. %s, evolution of $p\prime\prime$'%label)
        else:
            ax3.set_title(r'd) Exp. %s, evolution of $p$'%label)
        ax3.set_yticks(np.arange(5+1))
        ax3.set_xlabel('Time (h)')
        ax3.set_xlim([0,maxtime/3600])
        ax3.set_xticks(np.arange((maxtime+1)/3600))
    p.colorbar(c0,orientation='horizontal')
    p.colorbar(c1,orientation='horizontal')
    p.colorbar(c2,orientation='horizontal')
    #
    fig.savefig(filename,format='png',dpi=300)
    p.close(fig)

def plot_profiles(exp,filename):

    # Make plots
    fig, [ax1, ax2] = p.subplots(1,2,constrained_layout=True)
    fig.set_size_inches(6,3)

    which_cycle = 0
    tt1 = (which_cycle+1)*exp.da.assimilation_interval
    ax1 = plot_CBL_assimilation(exp,None,which_cycle=which_cycle,ax=ax1)
    ax1.set_ylim([0,2000])
    ax1.set_xlim([288,298])
    ax1.set_ylabel(r'height (m)')
    ax1.set_title(r'a) Analysis at $t=$%u s'%tt1)
    ax1.legend(frameon=False)

    which_cycle = 31
    tt2 = (which_cycle+1)*exp.da.assimilation_interval
    ax2 = plot_CBL_assimilation(exp,None,which_cycle=which_cycle,ax=ax2)
    ax2.set_ylim([0,2000])
    ax2.set_xlim([288,298])
    ax2.set_ylabel(r'height (m)')
    ax2.set_title(r'b) Analysis at $t=$%u s'%tt2)

    fig.savefig(filename,format='png',dpi=300)
    p.close(fig)