#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import numpy as np
from scipy.interpolate import CubicSpline
from scipy.special import erfinv,erf

verbose = True

def transform_none(x,kind):
    if kind=='dir':
        return x
    if kind=='inv':
        return x

def transform_log(x,kind):
    if kind=='dir':
        return np.log(x)
    if kind=='inv':
        return np.exp(x)

def transform_pit(x,kind):
    # mu and sigma could be guessed from the data, but...
    a = 0.5
    b = 4.5
    mu = 0.5*(a+b)
    sigma = (b-a)/12**0.5
    # takes a uniform, returns a normal
    if kind=='dir': 
        p = (x-a)/(b-a)
        xt = mu+sigma*2**0.5*erfinv(2*p-1)
        return xt
    # takes a normal), returns a uniform
    if kind=='inv':
        p = 0.5*(1+erf((x-mu)/sigma/2**0.5))
        xt = a+p*(b-a)
        return xt

class CBL:
    def __init__(self,settings):
        for k, v in settings.items():
            setattr(self, k, v)
        self.parameter_true = np.array([self.parameter_true])
        self.parameter_ensemble_min = np.array([self.parameter_ensemble_min])
        self.parameter_ensemble_max = np.array([self.parameter_ensemble_max])
        #self.parameter_inflation_rtps_alpha = np.array([self.parameter_inflation_rtps_alpha])
        if not hasattr(self,'do_parameter_estimation'):
            self.do_parameter_estimation == False
        if self.do_parameter_estimation:
            if self.parameter_transform == 'pit':
                self.parameter_transform = [ transform_pit ]
            elif self.parameter_transform == 'log':
                self.parameter_transform = [ transform_log ]
            elif self.parameter_transform == 'none':
                self.parameter_transform = [ transform_none ]
        else:
            self.parameter_transform = [ transform_none ]

    def initialize(self,argin,coordinates=None):

        """
        if input is scalar, interpret it as an ensemble size
        if input is a 1D or 2D array, interpret it as an initial state
        """
        self.zf = np.linspace(0,self.ztop,int(self.ztop/self.dz)+1)
        self.zt = 0.5*(self.zf[:-1]+self.zf[1:])
        self.nz = self.zt.size
        self.dt = 0.25*self.dz*self.dz/self.Kmax

        # scalar input is nens
        if np.isscalar(argin):
            self.nens = argin
            assert self.nens > 0, f"ensemble size must be greater than zero"

            # Define base state profile
            theta_init = self.theta_0+self.gamma*self.zt
            # For forced convection
            if self.is_bwind:
                u_init = np.zeros(self.nz)+self.ug
                v_init = np.zeros(self.nz)+self.vg
                x_init = np.zeros(self.nz*3)+np.nan
                x_init[:self.nz] = theta_init
                x_init[self.nz:self.nz*2] = u_init
                x_init[self.nz*2:self.nz*3] = v_init
                coordinates = np.zeros(self.nz*3)+np.nan
                coordinates[:self.nz] = self.zt
                coordinates[self.nz:self.nz*2] = self.zt
                coordinates[self.nz*2:self.nz*3] = self.zt
            else:
                x_init = theta_init
                coordinates = self.zt

            # If desired, append parameters to state vector;
            # initialize with the value in the middle of the admitted range
            # Coordinates of parameters are set to NaN
            # The reason is that parameters are assumed to be global.
            if self.do_parameter_estimation:
                if self.nens > 1:
                    for k in range(-self.parameter_number,0):
                        pmean = 0.5*(self.parameter_ensemble_min[k]+self.parameter_ensemble_max[k])
                        x_init = np.append(x_init,self.parameter_transform[k](pmean,kind='dir'))
                else:
                    for k in range(-self.parameter_number,0):
                        pmean = self.parameter_true[k]
                        x_init = np.append(x_init,self.parameter_transform[k](pmean,kind='dir'))
                coordinates = np.append(coordinates,self.parameter_ensemble_min*np.nan)

            # Replicate base state profile nens times
            x0 = np.tile(x_init,(self.nens,1)).T

        # 1D-array input is initial state of a deterministic run
        elif argin.ndim==1:
            self.nens = 1
            state_vector_length = self.nz
            if self.is_bwind:
                state_vector_length += 2*self.nz
            if self.do_parameter_estimation:
                state_vector_length += self.parameter_number
            assert argin.size == state_vector_length, f"1D input array should have size = nz*nvars+parameter_number, got {argin.shape[0]}"
            x0 = argin
            assert coordinates is not None, f"you must provide a valid coordinate array"
            assert x0.size == coordinates.size, f"coordinate vector must have same size as state vector"

        # 2D-array input is initial ensemble state
        elif argin.ndim==2:
            assert coordinates is not None, f"you must provide a valid coordinate array"
            state_vector_length = self.nz
            if self.is_bwind:
                state_vector_length += 2*self.nz
            if self.do_parameter_estimation:
                state_vector_length += self.parameter_number
            assert argin.shape[0] == state_vector_length, f"First dimension of 2D input array should be nz*nvars+parameter_number, got {argin.shape[0]}"
            self.nens = argin.shape[1]
            x0 = argin
            assert x0.shape[0] == coordinates.size, f"coordinate vector must have same size as state vector"

        # If desired, add ensemble perturbations to model state
        if self.nens>1 and self.perturb_ensemble_state:
            x0 = self.perturb_state(x0)

        # If desired, perturb parameters too
        # (last parameter_number elements of state vector)
        # Overwrite the pre-existing unperturbed values
        # This is needed for safety, because parameter transformations may be nonlinear
        if self.nens>1:
            #if self.do_parameter_estimation and hasattr(self,"initial_perturbed_parameters"):
            if hasattr(self,"initial_perturbed_parameters"):
                x0[-self.parameter_number:,:] = self.initial_perturbed_parameters
            else:
                if self.perturb_ensemble_parameters or self.do_parameter_estimation:
                    perturbed_parameters = self.create_parameter_perturbations()
                    self.initial_perturbed_parameters = perturbed_parameters
                if self.do_parameter_estimation:
                    assert (self.parameter_number==self.parameter_ensemble_min.size==self.parameter_ensemble_max.size),\
                        f'parameter min and max values must match parameters number,\
                        got {self.parameter_ensemble_min.size,self.parameter_ensemble_max.size,self.parameter_number}'
                    for k in range(-self.parameter_number,0):
                        assert (self.parameter_ensemble_min[k]<=self.parameter_ensemble_max[k]),\
                            f'parameter min must be less or equal to max,\
                            got {self.parameter_ensemble_min[k],self.parameter_ensemble_max[k]}'
                    x0[-self.parameter_number:,:] = self.initial_perturbed_parameters

        # Squeeze ensemble dimension if needed
        if self.nens==1:
            x0 = x0.squeeze()

        # Store initial state
        self.x0 = x0
        self.state_coordinates = coordinates
        
    def update(self,x0):
        self.x0 = x0

    def run(self,output_full_history=False):
        self.nt = int(self.maxtime//self.dt)+1

        if self.nens == 1:
            if verbose: print('#',end='')
            if output_full_history:
                self.x,self.history = self.cbl_model(self.x0,output_full_history=True)
            else:
                self.x = self.cbl_model(self.x0)
            if verbose: print(' ...done')
        else:
            self.x = np.zeros(self.x0.shape) + np.nan

            if self.perturb_ensemble_parameters and not self.do_parameter_estimation:
                pp = self.initial_perturbed_parameters
            else:
                pp = np.ones((1,self.nens))*np.nan

            if output_full_history:
                self.history = {}
                for n in range(self.nens):
                    if verbose: print('#',end='')
                    self.x[:,n],self.history['%04u'%n] = self.cbl_model(self.x0[:,n],perturbed_parameters=pp[:,n],output_full_history=True)
            else:
                for n in range(self.nens):
                    if verbose: print('#',end='')
                    self.x[:,n] = self.cbl_model(self.x0[:,n],perturbed_parameters=pp[:,n])
            if verbose: print(' ...done')

    def create_parameter_perturbations(self):
        # Initialize array to store perturbations
        pp = np.zeros((self.parameter_number,self.nens))

        for k in range(-self.parameter_number,0):
            RNG = np.random.default_rng(seed=self.rnseed)
            dum = RNG.uniform(self.parameter_ensemble_min[k], self.parameter_ensemble_max[k], size=self.nens)
            pp[k,:] = self.parameter_transform[k](dum , kind='dir')

        return pp

    def perturb_state(self,x0):
        # Random or uniform perturbations perturbations
        # differ only by magnitude of random sample
        if self.perturbations_type == "random":
            randomsize = self.nz
        elif self.perturbations_type == "uniform":
            randomsize = 1
        
        if self.perturbations_type == "random" or self.perturbations_type == "uniform":
            RNG = np.random.default_rng(seed=self.rnseed)
            ppt = RNG.standard_normal((randomsize,self.nens))
            if self.is_bwind:
                RNG = np.random.default_rng(seed=self.rnseed)
                ppu = RNG.standard_normal((self.nens,randomsize))
                ppv = RNG.standard_normal((self.nens,randomsize))

        # Smooth perturbations are slightly more complicated
        if self.perturbations_type == "smooth":

            # Define positions of perturbations
            # enforce two at boundaries to avoid extrapolation
            ipert = np.arange(0,self.perturbations_smooth_number-1)*(self.nz//(self.perturbations_smooth_number))
            ipert = np.append(ipert,self.nz-1)

            # Draw random perturbations, then interpolate
            randomsize = self.perturbations_smooth_number
            ppt = np.zeros((self.nz,self.nens))+np.nan
            RNG = np.random.default_rng(seed=self.rnseed)
            pert_t = RNG.standard_normal((randomsize,self.nens))
            for n in range(self.nens):
                f = CubicSpline(ipert,pert_t[:,n])
                ppt[:,n] = f(np.arange(self.nz))
            if self.is_bwind:
                ppu = np.zeros((self.nz,self.nens))+np.nan
                ppv = np.zeros((self.nz,self.nens))+np.nan
                RNG = np.random.default_rng(seed=self.rnseed)
                pert_u = RNG.standard_normal((randomsize,self.nens))
                pert_v = RNG.standard_normal((randomsize,self.nens))
                for n in range(self.nens):
                    f = CubicSpline(ipert,pert_u[:,n])
                    ppu[:,n] = f(np.arange(self.nz))
                    f = CubicSpline(ipert,pert_v[:,n])
                    ppv[:,n] = f(np.arange(self.nz)) 

        if self.perturbations_symmetric:
            # if n is odd, the last member remains unperturbed
            for n in range(self.nens//2):
                x0[:self.nz,n] += self.perturbations_theta_amplitude*ppt[:,n]
                x0[:self.nz,self.nens//2+n] -= self.perturbations_theta_amplitude*ppt[:,n]
                if self.is_bwind:
                    x0[self.nz:self.nz*2,n] += self.perturbations_uv_amplitude*ppu[:,n]
                    x0[self.nz:self.nz*2,self.nens//2+n] -= self.perturbations_uv_amplitude*ppu[:,n]
                    x0[self.nz*2:self.nz*3,n] += self.perturbations_uv_amplitude*ppv[:,n]
                    x0[self.nz*2:self.nz*3,self.nens//2+n] -= self.perturbations_uv_amplitude*ppv[:,n]
        else:
            for n in range(self.nens):
                x0[:self.nz,n] += self.perturbations_theta_amplitude*ppt[:,n]
                if self.is_bwind:
                    x0[self.nz:self.nz*2,n] += self.perturbations_uv_amplitude*ppu[:,n]
                    x0[self.nz*2:self.nz*3,n] += self.perturbations_uv_amplitude*ppv[:,n]
        return x0

    def cbl_model(self,x0,output_full_history=False,perturbed_parameters=np.nan):
        # Read settings in
        nt = self.nt
        dt = self.dt
        nz = self.nz
        dz = self.dz
        zf = self.zf
        zt = self.zt
        Kmax = self.Kmax
        Kmin = self.Kmin
        g = self.g
        f = self.f
        kvonk = self.kvonk
        z0 = self.z0
        theta_0 = self.theta_0
        Hmax = self.Hmax
        H0_perturbation_ampl_init = self.H0_perturbation_ampl_init
        H0_perturbation_ampl_time = self.H0_perturbation_ampl_time
        exct = self.exct
        ug = self.ug
        vg = self.vg
        pfac = self.pfac
        is_bwind = self.is_bwind
        is_cgrad = self.is_cgrad
        is_cycle = self.is_cycle
        parameter_transform = self.parameter_transform
        if hasattr(self,'do_parameter_estimation'):
            do_parameter_estimation = self.do_parameter_estimation
            parameter_number = self.parameter_number
        else:
            do_parameter_estimation = False
            parameter_number = 0

        # Initialize values and preallocate empty arrays
        xb = np.zeros(x0.shape)+np.nan
        theta = np.zeros((nz,nt))+np.nan
        H = np.zeros((nz+1,nt))+np.nan

        # For forced convection
        if is_bwind:
            u = np.zeros((nz,nt))+np.nan
            v = np.zeros((nz,nt))+np.nan
            M = np.zeros((nz+1,nt))+np.nan
            N = np.zeros((nz+1,nt))+np.nan

        # Sensible first guesses for PBL depth and eddy diffusivity
        h = dz*np.ones(nt)
        K = np.ones((nz+1,nt))*Kmin

        # Optionally, read perturbed parameters
        # This is not effective in the case of parameter estimation
        # The check is bypassed if the pfac vector is unset (=np.nan)
        if not isinstance(perturbed_parameters,np.ndarray):
            pass
        elif not do_parameter_estimation and ~np.isnan(perturbed_parameters.any()):
            pfac = parameter_transform[0](perturbed_parameters[0],kind='inv')

        # Unpack state vector
        theta[:,0] = x0[:nz]
        if is_bwind:
            u[:,0] = x0[nz:2*nz]
            v[:,0] = x0[2*nz:3*nz]
        if do_parameter_estimation:
            parameters = x0[-parameter_number:]
            pfac = parameter_transform[0](parameters[0],kind='inv')

        # Store initial state in temporary variable
        thetap = theta[:,0]
        if is_bwind:
            up = u[:,0]
            vp = v[:,0]

        # Integrate forward in time (ok for diffusion equation!)
        for j in range(1,nt):
            
            # Find model level nearest to h (parcel method)
            thr = thetap[0]+exct
            theta_profile = thetap[:]
            kh = theta_profile[theta_profile<=thr].size-1
            
            # Interpolate h
            h[j] = np.maximum(np.interp(thr,thetap[kh:kh+2],zt[kh:kh+2]),0.)

            # Set the surface sensible heat flux (H0)
            if is_cycle:
                H0 = Hmax*np.sin(2.*np.pi/86400.*dt*(j-1))
            else:
                H0 = Hmax
            
            # Add random perturbations to the initial value
            # Change the seed every time you go through this,
            # otherwise the perturbations are always the same
            seed = self.rnseed+j*200000
            RNG = np.random.default_rng(seed=seed)
            H0 += RNG.normal(scale=H0_perturbation_ampl_init)

            # Set the surface momentum flux (ustar)
            ustar = 0

            # For forced convection:
            # - you need surface momentum fluxes to integrate the momentum equations
            # - therefore ustar is not zero 
            if is_bwind:
                Cd     = (kvonk/np.log(zt[1]/z0))**2
                spd    = (up**2+vp**2)**0.5
                upwp   = -Cd*up[0]*spd[0]
                vpwp   = -Cd*vp[0]*spd[0]
                ustar  = (upwp**2+vpwp**2)**0.25
                # Iteration
                Cd_iterative = False
                if Cd_iterative:
                    # First guess
                    Cd     = (kvonk/np.log(zt[1]/z0))**2
                    delta  = 1e6
                    Cdmin = 1e-5
                    while delta>1e-6 and Cd>Cdmin:
                        upwp   = -Cd*up[0]*spd[0]
                        vpwp   = -Cd*vp[0]*spd[0]
                        ustar  = (upwp**2+vpwp**2)**0.25
                        zeta   = -kvonk*zf[1]*g*H0/(theta_0*ustar**3)
                        xx     = (1-15*zeta)**0.25
                        psim   = np.log((1+xx**2)/2)+2*np.log((1+xx)/2)-2*np.arctan(xx)+np.pi/2.
                        psim   = np.minimum(500,psim)
                        dummy  = (kvonk/(np.log(zf[1]/z0)-psim))**2
                        delta  = dummy-Cd
                    Cd  = np.maximum(Cdmin,dummy)

            # Define the eddy viscosity profile
            wstar  = (g/theta_0*h[j]*np.maximum(H0,0.001))**0.3333
            ws     = (ustar**3+0.7*kvonk*wstar**3)**0.3333
            zz     = (zf<h[j-1])*zf
            K[:,j] = np.minimum(kvonk*ws*zz*(1-zz/h[j-1])**pfac,Kmax)

            # Determine the countergradient correction
            if is_cgrad:
                gammac = 6.5*H0/ws/h[j]
            else:
                gammac = 0
            
            # Inverse grid spacing
            rdz = 1./self.dz

            # Add time-dependent surface perturbations
            # Change the seed every time you go through this,
            # otherwise the perturbations are always the same
            seed = self.rnseed+j*200000
            RNG = np.random.default_rng(seed=seed)

            # Then compute sensible heat flux and integrate T equation
            H[0,j]     = H0 + RNG.normal(scale=H0_perturbation_ampl_time)
            H[1:-1,j]  = -K[1:-1,j]*( (thetap[1:]-thetap[:-1])*rdz - gammac)
            H[nz,j]    = 2*H[nz-1,j]-H[nz-2,j]
            theta[:,j] = thetap[:]-dt*rdz*(H[1:,j]-H[:-1,j])
            thetap     = theta[:,j]

            # For forced convection
            if is_bwind:
                # Compute U momentum flux and integrate U equation
                M[0,j]    = upwp
                M[1:-1,j] = -K[1:-1,j]*(up[1:]-up[:-1])*rdz
                M[nz,j]   = 2*M[nz-1,j]-M[nz-2,j]
                u[:,j]    = up[:]-dt*rdz*(M[1:,j]-M[:-1,j])+dt*f*(vp[:]-vg)
                up        = u[:,j]
            
                # Compute V momentum flux and integrate V equation
                N[0,j]    = vpwp
                N[1:-1,j] = -K[1:-1,j]*(vp[1:]-vp[:-1])*rdz
                N[nz,j]   = 2*N[nz-1,j]-N[nz-2,j]
                v[:,j]    = vp[:]-dt*rdz*(N[1:,j]-N[:-1,j])-dt*f*(up[:]-ug)
                vp        = v[:,j]

        # Form state vector
        xb[:nz] = theta[:,-1]
        if is_bwind:
            xb[nz:2*nz] = u[:,-1]
            xb[2*nz:3*nz] = v[:,-1]
        if do_parameter_estimation:
            parameters = parameter_transform[0](pfac,kind='dir')
            xb[-parameter_number:] = parameters

        # Return state vector and, optionally, history of variables
        if is_bwind:
            history={'time' : dt*np.arange(0,nt), 'theta' : theta, 'u' : u, 'v': v}
        else:
            history={'time' : dt*np.arange(0,nt), 'theta' : theta}
        if output_full_history:
            return xb,history
        else:
            return xb