#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import numpy as np
from matplotlib import pyplot as p
import glob
import pandas as pd

is_sorted = lambda a: np.all(a[:-1] <= a[1:])

def decode_observations(observation_files):

    # Read set of CSV files into a pandas dataframe.
    # Based on:
    # https://stackoverflow.com/questions/20906474/import-multiple-csv-files-into-pandas-and-concatenate-into-one-dataframe    

    # Create a list of dataframes, one per file
    all_files = sorted(glob.glob(observation_files))
    li = []
    for filename in all_files:
        df = pd.read_csv(filename, index_col=None, header=0)
        li.append(df)
    nassim = len(li)

    # Browse through all dataframes and get the maximum
    # number of observation records
    nobs = 0
    for i in li:
        nobs = max(i.shape[0],nobs)

    # Create observation and coordinate arrays
    observations = np.zeros((nassim,nobs))+np.nan
    ocoords = np.zeros((nassim,nobs))+np.nan

    # Reading from the dataframes and
    # populate observation and coordinate arrays
    for i in range(nassim):
        ocoords[i,:] = li[i].get("z (m)")
        observations[i,:] = li[i].get("theta (K)")

    return observations,ocoords

def random_sort_observations(observations,ocoords,truths=None):

    # Check that input arrays have consistent dimensions
    assert (observations.shape == ocoords.shape),\
        f'observation and coordinate arrays must have the same size'

    # Preallocate arrays
    nassim,nobs = observations.shape
    shuffled_obs = np.zeros(observations.shape)+np.nan
    shuffled_coords = np.zeros(observations.shape)+np.nan

    # Randomize arrays, differently at each time
    if truths is None:
        indices = np.arange(nobs)
        for i in range(nassim):
            np.random.shuffle(indices)
            shuffled_obs[i,:] = observations[i,indices]
            shuffled_coords[i,:] = ocoords[i,indices]
            
        return shuffled_obs,shuffled_coords

    else:
        # Check that truths array has consistent dimensions
        assert (truths.shape == ocoords.shape),\
            f'observation and coordinate arrays must have the same size'

        # Preallocate truths array
        shuffled_truths = np.zeros(ocoords.shape)+np.nan

        # Randomize arrays, differently at each time
        indices = np.arange(nobs)
        for i in range(nassim):
            np.random.shuffle(indices)
            shuffled_obs[i,:] = observations[i,indices]
            shuffled_coords[i,:] = ocoords[i,indices]
            shuffled_truths[i,:] = truths[i,indices]
        return shuffled_obs,shuffled_coords,shuffled_truths

if __name__ == '__main__':

    printout = False

    # Read the observations
    observations,ocoords = decode_observations('./LES_data/Theta/*csv')

    if printout:
        ntim = 0
        for i in range(observations.shape[1]):
            print(ocoords[ntim,i],observations[ntim,i])

    # Sort the observations randomly
    observations,ocoords = random_sort_observations(observations,ocoords)
    if printout:
        for i in range(observations.shape[1]):
            print(ocoords[ntim,i],observations[ntim,i])

    makeplots = True
    # Make plots
    if makeplots:

        # If observations are shuffled, sort them again
        ntimes = ocoords.shape[0]
        for k in range(ntimes):
            if is_sorted(ocoords[k,:]):
                pass
            else:
                indices = np.argsort(ocoords[k,:])
                ocoords[k,:] = ocoords[k,indices]
                observations[k,:] = observations[k,indices]

        fig, ax1 = p.subplots(1,1,constrained_layout=True)
        fig.set_size_inches(4,3)

        tint = 300
        t = np.linspace(0,tint*(ocoords.shape[0]-1),ocoords.shape[0])[:,None]+np.zeros(ocoords.shape)
        t = t/3600.
        c = ax1.scatter(t,ocoords,s=5,c=observations,vmin=290,vmax=296)
        ax1.contour(t,ocoords,observations,
                np.linspace(290,290+0.003*2000,13),
                colors='black',
                linestyles='--',
                linewidths=0.75)
        ax1.set_ylim([0,2000])
        ax1.set_ylabel(r'Height (m)')
        ax1.set_xlabel(r'Time (s)')
        ax1.set_title(r'$\overline{\theta}$ (K) obs in t=(%u-%u) h'%(t.min(),t.max()))
        ax1.legend(frameon=False)
        p.colorbar(c,orientation='horizontal')
        fig.savefig('LES_obs.png',format='png',dpi=300)
        p.close(fig)
